"""Pyflakes support."""

import ast
from typing import Any, Dict, List
from pyflakes import checker

from pylama.lint import Linter as Abstract


# Path pyflakes messages
m = checker.messages
m.UnusedImport.message = f"W0611 {m.UnusedImport.message}"
m.RedefinedWhileUnused.message = f"W0404 {m.RedefinedWhileUnused.message}"
m.RedefinedInListComp.message = f"W0621 {m.RedefinedInListComp}"
m.ImportShadowedByLoopVar.message = f"W0621 {m.ImportShadowedByLoopVar}"
m.ImportStarUsed.message = f"W0401 {m.ImportStarUsed.message}"
m.ImportStarUsage.message = f"W0401 {m.ImportStarUsage.message}"
m.UndefinedName.message = f"E0602 {m.UndefinedName.message}"
m.DoctestSyntaxError.message = f"W0511 {m.DoctestSyntaxError.message}"
m.UndefinedExport.message = f"E0603 {m.UndefinedExport.message}"
m.UndefinedLocal.message = f"E0602 {m.UndefinedLocal.message}"
m.DuplicateArgument.message = f"E1122 {m.DuplicateArgument.message}"
m.LateFutureImport.message = f"W0410 {m.LateFutureImport.message}"
m.UnusedVariable.message = f"W0612 {m.UnusedVariable.message}"
m.ReturnWithArgsInsideGenerator.message = f"E0106 {m.ReturnWithArgsInsideGenerator.message}"
m.ReturnOutsideFunction.message = f"E0104 {m.ReturnOutsideFunction.message}"


class Linter(Abstract):
    """Pyflakes runner."""

    name = 'pyflakes'

    def run(self, path, *, code=None, params=None, **_) -> List[Dict[str, Any]]:  # noqa
        """Check code with pyflakes."""
        if params is None:
            params = {}

        builtins = params.get("builtins", "")
        if builtins:
            builtins = builtins.split(",")

        tree = compile(code, path, "exec", ast.PyCF_ONLY_AST)
        check = checker.Checker(tree, path, builtins=builtins)
        return [{
            'lnum': msg.lineno,
            'text': msg.message % msg.message_args,
            'type': msg.message[0]
        } for msg in check.messages]

#  pylama:ignore=E501,C0301



"""Support Vulture."""
from typing import Any, Dict, List

from vulture.core import Vulture, make_config, ERROR_CODES

from pylama.lint import Linter as BaseLinter


def parse_params(path, params=None):
    """Convert params from pylama."""
    return ['vulture'] + [f"--{key}={value}" for key, value in params.items()] + [path]


class Linter(BaseLinter):
    """vulture runner."""

    name = 'vulture'

    def run(self, path: str, *, params=None, **_) -> List[Dict[str, Any]]:  # noqa
        """Check code with vulture."""
        config = make_config(parse_params(path, params or {}))

        vulture = Vulture(
            verbose=config['verbose'],
            ignore_names=config['ignore_names'],
            ignore_decorators=config['ignore_decorators'],
        )
        vulture.scavenge([path], exclude=config['exclude'])
        unused_code_items = vulture.get_unused_code(
            min_confidence=config['min_confidence'],
            sort_by_size=config['sort_by_size']
        )
        result = []
        for item in unused_code_items:
            error_code = ERROR_CODES[item.typ]
            result.append({
                "lnum": item.first_lineno,
                "type": "V",
                "number": error_code,
                "text": f"{error_code} {item.message} ({item.confidence}% confidence)"
            })

        return result



#!/usr/bin/env python
# coding: utf-8
# (c) 2005 Divmod, Inc.  See LICENSE file for details


# commented code
#import os
# from foo import junk
# a = 3
a = 4
#foo(1, 2, 3)


class Message(object):
    message = ''
    message_args = ()
    def __init__(self, filename, loc, use_column=True):
        self.filename = filename
        self.lineno = loc.lineno
        self.col = getattr(loc, 'col_offset', None) if use_column else None
        test = 1
        if test == 1:
            if test == 1:
                return 28
            elif test == 2:
                return 28
            return 28
        elif test == 2:
            return 28

    def __str__(self):
        return '%s:%s: %s' % (self.filename, self.lineno, self.message % self.message_args)


class UnusedImport(Message):
    message = 'W402 %r imported but unused'

    def __init__(self, filename, lineno, name):
        Message.__init__(self, filename, lineno)
        self.message_args = (name,)


class RedefinedWhileUnused(Message):
    message = 'W801 redefinition of unused %r from line %r'

    def __init__(self, filename, lineno, name, orig_lineno):
        Message.__init__(self, filename, lineno)
        self.message_args = (name, orig_lineno)


class ImportShadowedByLoopVar(Message):
    message = 'W403 import %r from line %r shadowed by loop variable'

    def __init__(self, filename, lineno, name, orig_lineno):
        Message.__init__(self, filename, lineno)
        self.message_args = (name, orig_lineno)


class ImportStarUsed(Message):
    message = "W404 'from %s import *' used; unable to detect undefined names"

    def __init__(self, filename, lineno, modname):
        Message.__init__(self, filename, lineno)
        self.message_args = (modname,)


class UndefinedName(Message):
    message = 'W802 undefined name %r'

    def __init__(self, filename, lineno, name):
        Message.__init__(self, filename, lineno)
        self.message_args = (name,)


class UndefinedExport(Message):
    message = 'W803 undefined name %r in __all__'

    def __init__(self, filename, lineno, name):
        Message.__init__(self, filename, lineno)
        self.message_args = (name,)


class UndefinedLocal(Message):
    message = "W804 local variable %r (defined in enclosing scope on line " \
            "%r) referenced before assignment"

    def __init__(self, filename, lineno, name, orig_lineno):
        Message.__init__(self, filename, lineno)
        self.message_args = (name, orig_lineno)


class DuplicateArgument(Message):
    message = 'W805 duplicate argument %r in function definition'

    def __init__(self, filename, lineno, name):
        Message.__init__(self, filename, lineno)
        self.message_args = (name,)


class RedefinedFunction(Message):
    message = 'W806 redefinition of function %r from line %r'

    def __init__(self, filename, lineno, name, orig_lineno):
        Message.__init__(self, filename, lineno)
        self.message_args = (name, orig_lineno)


class LateFutureImport(Message):
    message = 'W405 future import(s) %r after other statements'

    def __init__(self, filename, lineno, names):
        Message.__init__(self, filename, lineno)
        self.message_args = (names,)


class UnusedVariable(Message):
    """
    Indicates that a variable has been explicitly assigned to but not actually
    used.
    """

    message = 'W806 local variable %r is assigned to but never used'

    def __init__(self, filename, lineno, names):
        Message.__init__(self, filename, lineno)
        self.message_args = (names,)
        error = 1 # noQa and some comments


class BadTyping(Message):
    """Test the MyPy linting."""

    message = 'error: No return value expected'

    def bad_method(self):  # type: () -> None
        """Return type mismatch."""
        return 1



"""pydocstyle support."""

from typing import Any, Dict, List

from pydocstyle import ConventionChecker as PyDocChecker
from pydocstyle.violations import conventions

from pylama.lint import Linter as Abstract


class Linter(Abstract):
    """Check pydocstyle errors."""

    name = "pydocstyle"

    def run(self, path: str, code: str = None, params=None, **_) -> List[Dict[str, Any]]:  # noqa
        """Check code with pydocstyle."""
        if params is None:
            params = {}

        convention_codes = conventions.get(params.get('convention'))

        return [
            {
                "lnum": e.line,
                # Remove colon after error code ("D403: ..." => "D403 ...").
                "text": (
                    e.message[0:4] + e.message[5:] if e.message[4] == ":" else e.message
                ),
                "type": "D",
                "number": e.code,
            }
            for e in PyDocChecker().check_source(
                code, path,
                params.get("ignore_decorators"),
                params.get("ignore_inline_noqa", False),
            ) if convention_codes is None or e.code in convention_codes
        ]



"""Code complexity checking."""
import ast
from typing import Dict, List, Any

from mccabe import McCabeChecker

from pylama.lint import Linter as Abstract


class Linter(Abstract):
    """Run complexity checking."""

    name = "mccabe"

    def run(self, path: str, *, code: str = None, params=None, **_) -> List[Dict[str, Any]]:  # noqa
        """Run Mccabe code checker."""
        if not code:
            return []

        if params is None:
            params = {}

        tree = compile(code, path, "exec", ast.PyCF_ONLY_AST)
        McCabeChecker.max_complexity = int(params.get("complexity", 10))
        return [
            {
                "lnum": lineno,
                "offset": offset,
                "text": text,
                "type": McCabeChecker._code,
            }
            for lineno, offset, text, _ in McCabeChecker(tree, path).run()
        ]


#  pylama:ignore=W0212



"""Pylama's shell support."""

import sys
import warnings
from os import path as op
from os import walk
from typing import List, Optional

from .check_async import check_async
from .config import CURDIR, Namespace, parse_options, setup_logger
from .core import LOGGER, run
from .errors import Error

DEFAULT_FORMAT = "{filename}:{lnum}:{col} [{etype}] {text}"
MESSAGE_FORMATS = {
    "pylint": "{filename}:{lnum}: [{etype}] {text}",
    "pycodestyle": "{filename}:{lnum}:{col} {text}",
    "parsable": DEFAULT_FORMAT,
}


def check_paths(
    paths: Optional[List[str]], options: Namespace, rootdir: str = None
) -> List[Error]:
    """Check the given paths.

    :param rootdir: Root directory (for making relative file paths)
    :param options: Parsed pylama options (from pylama.config.parse_options)
    """
    candidates = []
    for path in paths or options.paths:
        if not op.exists(path):
            continue

        if not op.isdir(path):
            candidates.append(op.abspath(path))

        for root, _, files in walk(path):
            candidates += [op.relpath(op.join(root, f), CURDIR) for f in files]

    if not candidates:
        return []

    if rootdir is None:
        path = candidates[0]
        rootdir = path if op.isdir(path) else op.dirname(path)

    linters = options.linters
    if not options.force:
        candidates = [
            path for path in candidates if any(l.allow(path) for _, l in linters)
        ]

    if options.concurrent:
        return check_async(candidates, options, rootdir)

    errors = []
    for path in candidates:
        errors += run(path=path, rootdir=rootdir, options=options)

    return errors


def check_path(
    options: Namespace,
    rootdir: str = None,
    candidates: List[str] = None,
    code: str = None,  # noqa
) -> List[Error]:
    """Support legacy code."""
    warnings.warn(
        "pylama.main.check_path is depricated and will be removed in pylama 9",
        DeprecationWarning,
    )
    return check_paths(candidates, options=options, rootdir=rootdir)


def shell(args=None, error=True):
    """Endpoint for console.

    Parse a command arguments, configuration files and run a checkers.

    :return list: list of errors
    :raise SystemExit:

    """
    if args is None:
        args = sys.argv[1:]

    options = parse_options(args)
    setup_logger(options)
    LOGGER.info(options)

    # Install VSC hook
    if options.hook:
        from .hook import install_hook  # noqa

        for path in options.paths:
            return install_hook(path)

    return process_paths(options, error=error)


def process_paths(
    options: Namespace, candidates: List[str] = None, error: bool = True
) -> List[Error]:
    """Process files and log errors."""
    errors = check_paths(candidates, options, rootdir=CURDIR)
    pattern = MESSAGE_FORMATS.get(options.format, DEFAULT_FORMAT)

    for err in errors:
        if options.abspath:
            err.filename = op.abspath(err.filename)
        LOGGER.warning(
            pattern.format(
                filename=err.filename,
                lnum=err.lnum,
                col=err.col,
                text=err.text,
                etype=err.type,
            )
        )

    if error:
        sys.exit(int(bool(errors)))

    return errors


if __name__ == "__main__":
    shell()

# pylama:ignore=F0001



"""Support Radon."""

from typing import Any, Dict, List
from radon.visitors import ComplexityVisitor
from radon.complexity import add_inner_blocks

from pylama.lint import Linter as Abstract


class Linter(Abstract):
    """Radon runner."""

    name = 'radon'

    def run(self, path: str, code=None, params=None, **_) -> List[Dict[str, Any]]:  # noqa  # noqa
        """Check code with Radon."""
        if params is None:
            params = {}

        complexity = params.get('complexity', 10)
        no_assert = params.get('no_assert', False)
        show_closures = params.get('show_closures', False)

        visitor = ComplexityVisitor.from_code(code, no_assert=no_assert)
        blocks = visitor.blocks
        if show_closures:
            blocks = add_inner_blocks(blocks)

        return [
            {'lnum': block.lineno, 'col': block.col_offset, 'type': 'R', 'number': 'R901',
             'text': f"R901 {block.name} is too complex {block.complexity}"}
            for block in visitor.blocks if block.complexity > complexity
        ]



"""Commented-out code checking."""

from typing import List, Dict, Any

from eradicate import Eradicator
from pylama.lint import Linter as Abstract


class Linter(Abstract):
    """Run commented-out code checking."""

    name = 'eradicate'

    def run(self, path: str, *, code: str = None, **_) -> List[Dict[str, Any]]:  # noqa
        """Eradicate code checking.

        TODO: Support params
        """
        if not code:
            return []

        eradicator = Eradicator()
        line_numbers = eradicator.commented_out_code_line_numbers(code)
        lines = code.split('\n')

        result = []
        for line_number in line_numbers:
            line = lines[line_number - 1]
            result.append(dict(
                lnum=line_number,
                offset=len(line) - len(line.rstrip()),
                text=str('E800 Found commented out code: ') + line,
                type='E800',
            ))
        return result



"""Custom module loader."""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from importlib import import_module
from pathlib import Path
from pkgutil import walk_packages
from typing import Any, Dict, List, Optional, Type

from pkg_resources import iter_entry_points

LINTERS: Dict[str, Type[Linter]] = {}


class Linter(metaclass=ABCMeta):
    """Abstract class for linter plugin."""

    name: Optional[str] = None

    def __init_subclass__(cls) -> None:
        """Register subclasses."""
        if cls.name is not None:
            LINTERS[cls.name] = cls

    @staticmethod
    def allow(path: str) -> bool:
        """Check path is relevant for linter."""
        return path.endswith(".py")

    @abstractmethod
    def run(self, path: str, **meta) -> List[Dict[str, Any]]:
        """Not implemented."""
        raise NotImplementedError(__doc__)


# Import default linters
for _, name, _ in walk_packages([str(Path(__file__).parent)]):  # type: ignore
    try:
        import_module(f"{__name__}.{name}")
    except ImportError:
        pass

# Import installed linters
for entry in iter_entry_points("pylama.linter"):
    if entry.name not in LINTERS:
        try:
            LINTERS[entry.name] = entry.load()
        except ImportError:
            pass



""" Don't duplicate same errors from different linters. """
from __future__ import annotations

import re
from collections import defaultdict
from typing import DefaultDict, Generator, List, Set


PATTERN_NUMBER = re.compile(r'^[A-Z]\d+$')

DUPLICATES = {key: values for values in (

    # multiple statements on one line
    {('pycodestyle', 'E701'), ('pylint', 'C0321')},

    # unused variable
    {('pylint', 'W0612'), ('pyflakes', 'W0612')},

    # undefined variable
    {('pylint', 'E0602'), ('pyflakes', 'E0602')},

    # unused import
    {('pylint', 'W0611'), ('pyflakes', 'W0611')},

    # whitespace before ')'
    {('pylint', 'C0326'), ('pycodestyle', 'E202')},

    # whitespace before '('
    {('pylint', 'C0326'), ('pycodestyle', 'E211')},

    # multiple spaces after operator
    {('pylint', 'C0326'), ('pycodestyle', 'E222')},

    # missing whitespace around operator
    {('pylint', 'C0326'), ('pycodestyle', 'E225')},

    # unexpected spaces
    {('pylint', 'C0326'), ('pycodestyle', 'E251')},

    # long lines
    {('pylint', 'C0301'), ('pycodestyle', 'E501')},

    # statement ends with a semicolon
    {('pylint', 'W0301'), ('pycodestyle', 'E703')},

    # multiple statements on one line
    {('pylint', 'C0321'), ('pycodestyle', 'E702')},

    # bad indentation
    {('pylint', 'W0311'), ('pycodestyle', 'E111')},

    # wildcart import
    {('pylint', 'W00401'), ('pyflakes', 'W0401')},

    # module docstring
    {('pydocstyle', 'D100'), ('pylint', 'C0111')},

) for key in values}


def remove_duplicates(errors: List[Error]) -> Generator[Error, None, None]:
    """ Filter duplicates from given error's list. """
    passed: DefaultDict[str, Set] = defaultdict(set)
    for error in errors:
        key = error.linter, error.number
        if key in DUPLICATES:
            if key in passed[error.lnum]:
                continue
            passed[error.lnum] = DUPLICATES[key]
        yield error


class Error:

    """ Store an error's information. """

    __slots__ = 'linter', 'col', 'lnum', 'type', 'text', 'filename', 'number'

    def __init__(self, linter="", col=1, lnum=1, type="E",
                 text="unknown error", filename="", number="", **_):
        """ Init error information with default values. """
        text = str(text).strip().replace('\n', ' ')
        if linter:
            text = f"{text} [{linter}]"
        number = number or text.split(' ', 1)[0]
        if not PATTERN_NUMBER.match(number):
            number = ""

        self.linter = linter
        self.col = col
        self.lnum = int(lnum)
        self.type = type[:1]
        self.text = text
        self.filename = filename
        self.number = number

    def __repr__(self):
        return f"<Error: {self.number} {self.linter}>"

# pylama:ignore=W0622,D,R0924



"""Parse arguments from command line and configuration files."""
import fnmatch
import logging
import os
import re
import sys
from argparse import ArgumentParser, Namespace
from typing import Any, Collection, List, Optional, Set, Tuple, Type, Union

from . import __version__
from .libs import inirama
from .lint import LINTERS, Linter

#: A default checkers
DEFAULT_LINTERS = "pycodestyle", "pyflakes", "mccabe"

CURDIR = os.getcwd()
CONFIG_FILES = "pylama.ini", "setup.cfg", "tox.ini", "pytest.ini"

#: The skip pattern
SKIP_PATTERN = re.compile(r"# *noqa\b", re.I).search

# Parse a modelines
MODELINE_RE = re.compile(
    r"^\s*#\s+(?:pylama:)\s*((?:[\w_]*=[^:\n\s]+:?)+)", re.I | re.M
)

# Setup a logger
LOGGER = logging.getLogger("pylama")
LOGGER.propagate = False
STREAM = logging.StreamHandler(sys.stdout)
LOGGER.addHandler(STREAM)


class _Default:
    def __init__(self, value=None):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"<_Default [{self.value}]>"


def split_csp_str(val: Union[Collection[str], str]) -> Set[str]:
    """Split comma separated string into unique values, keeping their order."""
    if isinstance(val, str):
        val = val.strip().split(",")
    return set(x for x in val if x)


def parse_linters(linters: str) -> List[Tuple[str, Type[Linter]]]:
    """Initialize choosen linters."""
    result = []
    for name in split_csp_str(linters):
        linter = LINTERS.get(name)
        if linter:
            result.append((name, linter))
        else:
            logging.warning("Linter `%s` not found.", name)
    return result


def get_default_config_file(rootdir: str = None) -> Optional[str]:
    """Search for configuration file."""
    if rootdir is None:
        return DEFAULT_CONFIG_FILE

    for path in CONFIG_FILES:
        path = os.path.join(rootdir, path)
        if os.path.isfile(path) and os.access(path, os.R_OK):
            return path

    return None


DEFAULT_CONFIG_FILE = get_default_config_file(CURDIR)


PARSER = ArgumentParser(description="Code audit tool for python.")
PARSER.add_argument(
    "paths",
    nargs="*",
    default=_Default([CURDIR]),
    help="Paths to files or directories for code check.",
)

PARSER.add_argument("--verbose", "-v", action="store_true", help="Verbose mode.")

PARSER.add_argument("--version", action="version", version="%(prog)s " + __version__)

PARSER.add_argument(
    "--format",
    "-f",
    default=_Default("pycodestyle"),
    choices=["pep8", "pycodestyle", "pylint", "parsable"],
    help="Choose errors format (pycodestyle, pylint, parsable).",
)

PARSER.add_argument(
    "--select",
    "-s",
    default=_Default(""),
    type=split_csp_str,
    help="Select errors and warnings. (comma-separated list)",
)

PARSER.add_argument(
    "--sort",
    default=_Default(""),
    type=split_csp_str,
    help="Sort result by error types. Ex. E,W,D",
)

PARSER.add_argument(
    "--linters",
    "-l",
    default=_Default(",".join(DEFAULT_LINTERS)),
    type=parse_linters,
    help=(
        f"Select linters. (comma-separated). Choices are {','.join(s for s in LINTERS)}."
    ),
)

PARSER.add_argument(
    "--ignore",
    "-i",
    default=_Default(""),
    type=split_csp_str,
    help="Ignore errors and warnings. (comma-separated)",
)

PARSER.add_argument(
    "--skip",
    default=_Default(""),
    type=lambda s: [re.compile(fnmatch.translate(p)) for p in s.split(",") if p],
    help="Skip files by masks (comma-separated, Ex. */messages.py)",
)

PARSER.add_argument("--report", "-r", help="Send report to file [REPORT]")
PARSER.add_argument("--hook", action="store_true", help="Install Git (Mercurial) hook.")

PARSER.add_argument(
    "--concurrent",
    "--async",
    action="store_true",
    help="Enable async mode. Useful for checking a lot of files. ",
)

PARSER.add_argument(
    "--options",
    "-o",
    default=DEFAULT_CONFIG_FILE,
    metavar="FILE",
    help=(
        "Specify configuration file. "
        f"Looks for {', '.join(CONFIG_FILES[:-1])}, or {CONFIG_FILES[-1]}"
        f" in the current directory (default: {DEFAULT_CONFIG_FILE})"
    ),
)

PARSER.add_argument(
    "--force",
    "-F",
    action="store_true",
    default=_Default(False),
    help="Force code checking (if linter doesn't allow)",
)

PARSER.add_argument(
    "--abspath",
    "-a",
    action="store_true",
    default=_Default(False),
    help="Use absolute paths in output.",
)


ACTIONS = dict((a.dest, a) for a in PARSER._actions)  # pylint: disable=protected-access


def parse_options(  # noqa
    args: List[str] = None, config: bool = True, rootdir: str = CURDIR, **overrides
) -> Namespace:
    """Parse options from command line and configuration files."""
    args = args or []

    # Parse args from command string
    options = PARSER.parse_args(args)
    options.file_params = {}
    options.linters_params = {}

    # Compile options from ini
    if config:
        cfg = get_config(str(options.options), rootdir=rootdir)
        for opt, val in cfg.default.items():
            LOGGER.info("Find option %s (%s)", opt, val)
            passed_value = getattr(options, opt, _Default())
            if isinstance(passed_value, _Default):
                if opt == "paths":
                    val = val.split()
                if opt == "skip":
                    val = fix_pathname_sep(val)
                setattr(options, opt, _Default(val))

        # Parse file related options
        for name, opts in cfg.sections.items():

            if name == cfg.default_section:
                continue

            if name.startswith("pylama"):
                name = name[7:]

            if name in LINTERS:
                options.linters_params[name] = dict(opts)
                continue

            mask = re.compile(fnmatch.translate(fix_pathname_sep(name)))
            options.file_params[mask] = dict(opts)

    # Override options
    for opt, val in overrides.items():
        setattr(options, opt, process_value(opt, val))

    # Postprocess options
    for name in options.__dict__:
        value = getattr(options, name)
        if isinstance(value, _Default):
            setattr(options, name, process_value(name, value.value))

    if options.concurrent and "pylint" in options.linters:
        LOGGER.warning("Can't parse code asynchronously with pylint enabled.")
        options.concurrent = False

    return options


def process_value(name: str, value: Any) -> Any:
    """Compile option value."""
    action = ACTIONS.get(name)
    if not action:
        return value

    if callable(action.type):
        return action.type(value)

    if action.const:
        return bool(int(value))

    return value


def get_config(ini_path: str = None, rootdir: str = None) -> inirama.Namespace:
    """Load configuration from INI."""
    config = inirama.Namespace()
    config.default_section = "pylama"

    if not ini_path:
        path = get_default_config_file(rootdir)
        if path:
            config.read(path)
    else:
        config.read(ini_path)

    return config


def setup_logger(options: Namespace):
    """Do the logger setup with options."""
    LOGGER.setLevel(logging.INFO if options.verbose else logging.WARN)
    if options.report:
        LOGGER.removeHandler(STREAM)
        LOGGER.addHandler(logging.FileHandler(options.report, mode="w"))

    if options.options:
        LOGGER.info("Try to read configuration from: %r", options.options)


def fix_pathname_sep(val: str) -> str:
    """Fix pathnames for Win."""
    return val.replace(os.altsep or "\\", os.sep)


# pylama:ignore=W0212,D210,F0001



"""Pylama's core functionality.

Prepare params, check a modeline and run the checkers.
"""
import os.path as op
from typing import Any, Collection, Dict, Generator, List, Optional, Set, Tuple

from .config import (CURDIR, LOGGER, MODELINE_RE, SKIP_PATTERN, Namespace,
                     process_value)
from .errors import Error, remove_duplicates
from .lint import LINTERS


def run(path: str, rootdir: str = CURDIR, options: Namespace = None) -> List[Error]:
    """Run code checkers with given params.

    :param path: (str) A file's path.
    """
    code = None
    errors: List[Error] = []
    fileconfig = {}
    linters_params = {}
    params = {}
    path = op.relpath(path, rootdir)

    if options:
        linters_params = options.linters_params
        for mask in options.file_params:
            if mask.match(path):
                fileconfig.update(options.file_params[mask])

        if options.skip and any(p.match(path) for p in options.skip):
            LOGGER.info("Skip checking for path: %s", path)
            return []

    try:
        with open(path, encoding="utf-8") as source:
            code = source.read()
            params = build_params(options, fileconfig, parse_modeline(code))
            if params.get("skip"):
                LOGGER.info("Skip: %s", path)
                return errors

            LOGGER.debug("Checking params: %s", params)
            linters_to_run = get_linters(path, params, linters_params)
            for linter_cls, lname, select, ignore, lparams in linters_to_run:
                LOGGER.info("Run [%s] %s - %s", lname, path, lparams)
                linter_errors = linter_cls().run(
                    path, code=code, ignore=ignore, select=select, params=lparams
                )
                if linter_errors:
                    errors += filter_errors(
                        [
                            Error(filename=path, linter=lname, **er)
                            for er in linter_errors
                        ],
                        select=select,
                        ignore=ignore,
                    )

    except IOError as exc:
        LOGGER.error("IOError %s", exc)
        errors.append(Error(text=str(exc), filename=path, linter="pylama"))

    except SyntaxError as exc:
        LOGGER.error("SyntaxError %s", exc)
        errors.append(
            Error(
                linter="pylama",
                lnum=exc.lineno,
                col=exc.offset,
                text=f"E0100 SyntaxError: {exc.args[0]}",
                filename=path,
            )
        )

    except Exception as exc:  # noqa
        LOGGER.exception(exc)

    errors = list(remove_duplicates(errors))

    if code and errors:
        errors = filter_skiplines(code, errors)

    sorter = default_sorter
    if options and options.sort:
        sort = dict((v, n) for n, v in enumerate(options.sort, 1))
        sorter = lambda err: (sort.get(err.type, 999), err.lnum)

    return sorted(errors, key=sorter)


def get_linters(
    path: str, params: Dict[str, Any], linters_params: Dict[str, Any]
) -> Generator[Tuple, None, None]:
    """Prepare linters and params."""
    linters = params.get("linters") or LINTERS
    for item in linters:
        if not isinstance(item, tuple):
            item = item, LINTERS.get(item)

        lname, lclass = item
        if not (lclass and lclass.allow(path)):  # noqa
            continue

        lparams = linters_params.get(lname, {})
        ignore, select = merge_params(params, lparams)
        yield (lclass, lname, select, ignore, lparams)


def parse_modeline(code: str) -> Dict[str, str]:
    """Parse params from file's modeline."""
    seek = MODELINE_RE.search(code)
    if seek:
        return dict(v.split("=", 1) for v in seek.group(1).split(":"))  # type: ignore

    return {}


def build_params(
    options: Optional[Namespace], *configs: Dict[str, str]
) -> Dict[str, Any]:
    """Prepare and merge a params from modelines and configs."""
    if options:
        params: Dict[str, Any] = dict(
            skip=False,
            linters=options.linters,
            ignore=options.ignore,
            select=options.select,
        )
    else:
        params = dict(skip=False, ignore=set(), select=set(), linters=[])

    for config in configs:
        if not config:
            continue

        for key in ("ignore", "select"):
            if key in config:
                params[key] |= process_value(key, config[key])

        if "linters" in config:
            params["linters"] = process_value("linters", config["linters"])

        params["skip"] = bool(int(config.get("skip", False)))

    return params


def filter_errors(
    errors: List[Error], select: Collection[str] = None, ignore: Collection[str] = None
) -> Generator[Error, None, None]:
    """Filter errors by select and ignore options."""
    select = select or []
    ignore = ignore or []

    for err in errors:
        for rule in select:
            if err.number.startswith(rule):
                yield err
                break
        else:
            for rule in ignore:
                if err.number.startswith(rule):
                    break
            else:
                yield err


def filter_skiplines(code: str, errors: List[Error]) -> List[Error]:
    """Filter lines by `noqa`."""
    lnums = set(er.lnum for er in errors)
    skipped = set(
        lnum
        for lnum, line in enumerate(code.split("\n"), 1)
        if lnum in lnums and SKIP_PATTERN(line)
    )

    if skipped:
        errors = [er for er in errors if er.lnum not in skipped]

    return errors


def merge_params(
    params: Dict[str, Any], lparams: Dict[str, Any]
) -> Tuple[Set[str], Set[str]]:
    """Merge global ignore/select with linter local params."""
    ignore = params.get("ignore", set())
    if "ignore" in lparams:
        ignore = ignore | set(lparams["ignore"].split(","))

    select = params.get("select", set())
    if "select" in lparams:
        select = select | set(lparams["select"].split(","))

    return ignore, select


def default_sorter(err: Error) -> Any:
    """Sort by line number."""
    return err.lnum


# pylama:ignore=R0912,D210,F0001
